/**
   Бибилиотеки для прошивки - MySensors-1.5.4
   Скачать библиотеку: https://github.com/mysensors/Arduino/archive/1.5.4.zip
   
   Created by Henrik Ekblad <henrik.ekblad@mysensors.org>
   Copyright (C) 2013-2017 Sensnology AB
   Full contributor list: https://github.com/mysensors/Arduino/graphs/contributors

   ******************************
   Основной скетч и идею взял здесь, у Алексея Степановa @avstepanov
  
   https://geektimes.ru/post/258354/   Тестирование и подготовка
   https://geektimes.ru/post/258388/   Hardware выключателя
   https://geektimes.ru/post/258430/   Software выключателя
   https://geektimes.ru/post/258442/   «Центр управления»
    
   ******************************
   Версия прошивки "Multy-Switch-Relay"
   Version 2.3 - Ahelper

   Документация по прошивке
   "Умный выключатель 2"
   http://majordomo.smartliving.ru/forum/viewforum.php?f=20

   ******************************
*/
#include <MySigningNone.h>
#include <MyTransportNRF24.h>
#include <MyHwATMega328.h>
#include <MySensor.h>
#include <avr/wdt.h>
#include <SPI.h>
#include <Bounce2.h>

#define DEBUGS // Если нужен вывод отладочных сообщений - закомментировать

// Определим задержку для длинного нажатия
#define LONGPRESS 2000  // 2 секунды

// Define MySensor gateway
MySensor gw;

#define RADIO_ID 45 // Id ноды, независимо от того, какой канал вы назначили
#define CH 3  // Колличество каналов
#define Num 9 // Колличество передаваемых сообщений (датчиков) на шлюз

// Создадим структуру для хранения параметров "канала"
typedef struct {
  int button;                // пин кнопки
  int relay;                 // пин реле
  boolean state;             // Состояние реле (также сохраняется в EEPROM)
  unsigned long power_on;    // время, когда нагрузка была включена
  unsigned long auto_off;    // время, через которое нагрузку автоматически выключить
  unsigned long timer;       // MySwitch[0].timer - время до автоматического включения вытяжки (после включения света)
                             // MySwitch[1].timer - время до автоматического выключения вытяжки (после выключения света)
                             // MySwitch[2].timer - время не установлено
  unsigned long time_off;    // время, автовыключения 
  boolean autostate;         // флаг, означающий, что ждем автовыключение 
  unsigned long press_start; // время, когда кнопку нажали
  unsigned long press_stop;  // время, когда кнопку отпустили
}
Channel;

// Определим (установим) параметры каналов
// MySwitch[0]. - канал освещения ванной
// MySwitch[1]. - канал вытяжки ванной
// MySwitch[2]. - канал освещения прихожей
// button, relay, state, power_on, auto_off, timer, time_off, autostate, press_start, press_stop
Channel MySwitch[CH] = {
  6, 3, LOW, 0, 0, 0, 0, false, 0, 0,
  7, 4, LOW, 0, 0, 0, 0, false, 0, 0,
  8, 5, LOW, 0, 0, 0, 0, false, 0, 0
};

// Флаг для дополнительной логики
boolean logicFlag = false;
boolean onFlag = false;
boolean offFlag = false;

Bounce debouncer[CH];
MyMessage msg[Num];

// Функция установок
void setup()
{
  //wdt_disable(); // Бесполезная строка до которой не доходит выполнение при bootloop

  gw.begin(incomingMessage, RADIO_ID, true);  // Инициализация входящих сообщений от шлюза
  delay(250); // пауза 250 мс
  gw.sendSketchInfo("Multy-Switch-Relay", "2.3"); // Отправка версии скетча на шлюз
  delay(250); // пауза 250 мс

  // Инициализировать реле и соответствующие кнопки по каналам 0-2
  for (int i = 0; i < CH; i++)  {
    msg[i].sensor = i;  // Инициализировать сообщения реле
    msg[i].type = V_STATUS; // Установим тип данных, сообщений датчика
    debouncer[i] = Bounce();  // Инициализировать debouncer
    debouncer[i].attach(MySwitch[i].button);
    debouncer[i].interval(5);
    pinMode(MySwitch[i].button, INPUT_PULLUP);  // Установка входа кнопок и включение подтяжки
    pinMode(MySwitch[i].relay, OUTPUT); // Установка выхода реле
    gw.present(i, S_CUSTOM, "Switch"); // Представить тип датчика на шлюз
    MySwitch[i].state = gw.loadState(i); // Извлечь последние значения из EEPROM
    digitalWrite(MySwitch[i].relay, MySwitch[i].state); // И соответственно установить реле вкл/выкл
    gw.send(msg[i].set(MySwitch[i].state)); // Информировать шлюз о последнем статусе
    delay(250); // пауза 250 мс
  }
  // Инициализировать таймеры "auto_off" по каналам 3-5
  for (int i = 0; i < CH; i++)  {
    msg[i+CH].sensor = i+CH;  // Инициализировать сообщения таймеров "auto_off"
    msg[i+CH].type = V_WEIGHT; // Установим тип данных, сообщений датчика
    gw.present(i+CH, S_CUSTOM, "auto_off"); // Представить тип датчика на шлюз
    MySwitch[i].auto_off = gw.loadState(i+CH); // Извлечь последние значения из EEPROM
    gw.send(msg[i+CH].set(MySwitch[i].auto_off)); // Информировать шлюз об установленном времени в минутах
    delay(250); // пауза 250 мс
  }
  // Инициализировать таймеры "timer" по каналам 6-8
   for (int i = 0; i < CH; i++)  {
    msg[i+CH+CH].sensor = i+CH+CH;  // Инициализировать сообщения таймеров "timer"
    msg[i+CH+CH].type = V_FLOW; // Установим тип данных, сообщений датчика
    gw.present(i+CH+CH, S_CUSTOM, "timer"); // Представить тип датчика на шлюз
    MySwitch[i].timer = gw.loadState(i+CH+CH); // Извлечь последние значения из EEPROM
    gw.send(msg[i+CH+CH].set(MySwitch[i].timer)); // Информировать шлюз об установленном времени в минутах
    delay(250); // пауза 250 мс
  }
  
// Если закомментировано //#define DEBUGS, то отправляются сообщения в COM порт
#ifndef DEBUGS
  Serial.println("Ready!");
#endif
  
  //wdt_enable (WDTO_1S); // Устанавливаем значение ожидания сторожевой собаки 1 сек.
}

// Функция основного процесса
void loop() {
  gw.process();

  // Работаем с кнопками
  button_read();   // переход в функцию "button_read()"

  // Отработаем автовыключение
  autoOff();      // переход в функцию "autoOff()"

  // Проверим дополнительную логику работы
  chkLogic();     // переход в функцию "chkLogic()"

  // Сбросим сторожевую собаку
  //wdt_reset();
}

// Функция обработки нажатия кнопок
void  button_read(){
  // Цикл по всем каналам
for (byte i = 0; i < CH; i++)  {
    if ( debouncer[i].update() ) {
      if ( debouncer[i].read() == LOW) {
        // Фиксируем время старта нажатия
        MySwitch[i].press_start = millis();
      }
      else {
        // Переход в функцию "void pressDetect", с параметрами
        // Для определения, какое нажатие было (короткое или длинное)
        pressDetect(i, millis());
      }
    }
  }
}

// Функция реализации выключателя, отрабатывает строку вида "doSwitch(i, LOW);"
void doSwitch(int ch, boolean state){
  // Устанавливаем требуемое состояние выключателя
  MySwitch[ch].state = state;
  // Отправить состояние на шлюз
  gw.send(msg[ch].set(MySwitch[ch].state)); // Информировать шлюз о статусе
  gw.saveState( ch, MySwitch[ch].state ); // Сохранить состояние датчика в EEPROM (положение == номер датчика)
  // Если выключатель перешел в положение "ВКЛ"
  if (MySwitch[ch].state == HIGH) {
    // Зафиксируем время включения
    MySwitch[ch].power_on = millis();
    // И если время автоотключения больше нуля и не равно нулю
    if((MySwitch[ch].auto_off > 0) && (MySwitch[ch].auto_off != 0)) {
      // Посчитаем время, когда надо будет автоматически выключить
      MySwitch[ch].time_off = MySwitch[ch].power_on + MySwitch[ch].auto_off*60000;
      MySwitch[ch].autostate = true;  // Переведем флаг автовыключения в "true"
    }
#ifndef DEBUGS
    Serial.print("ON ch ");
    Serial.println(ch);
#endif
  }
  else {  // Если выключатель в положение "ВЫКЛ"
    // Отключим режим автовыключения
    MySwitch[ch].autostate = false;
    // Сбросим время автовыключения
    MySwitch[ch].time_off = 0;
#ifndef DEBUGS
    Serial.print("OFF ch ");
    Serial.println(ch);
#endif
  }
  // Установим выход реле в требуемое состояние
  digitalWrite(MySwitch[ch].relay,MySwitch[ch].state);
}
// Функция автовыключения
void autoOff(){
  // Цикл по всем каналам
  for (int i=0; i < CH; i++) {
    // Если время выключения подошло - щелкнем выключателем
    if ((millis() >= MySwitch[i].time_off) && MySwitch[i].autostate) {
      MySwitch[i].autostate = false;  // Отключим режим автовыключения
      doSwitch(i, LOW); // Переход в функцию "doSwitch", с параметрами
#ifndef DEBUGS
      Serial.print("Auto OFF ");
      Serial.println(i);
#endif
    }
  }
}

// Функция определения длинны нажатия на клавишу и выполняем действия
void pressDetect(int ch, unsigned long p_stop) {
  if  (MySwitch[ch].press_start != 0) {
    if (((p_stop-MySwitch[ch].press_start) < LONGPRESS) && (p_stop-MySwitch[ch].press_start) > 0) {
      // короткое нажатие
      MySwitch[ch].press_stop = p_stop;
#ifndef DEBUGS
      Serial.print("Short press ch ");
      Serial.println(ch);
#endif
      // Отправляем в функцию "doSwitch", выключатель "ch", состояние вкл/выкл
      doSwitch(ch, MySwitch[ch].state ? LOW : HIGH);
      }
    else {
      // длинное нажатие
#ifndef DEBUGS
      Serial.print("Long press ch ");
      Serial.println(ch);
#endif
      doSwitch(2, MySwitch[2].state ? LOW : HIGH);  // Отправляем состояние вкл/выкл выключателем прихожей 
      gw.saveState( 2, MySwitch[2].state ); // Сохранить состояние датчика в EEPROM (положение == номер датчика)
    }
  }
}

// Функция дополнительной логики работы
void chkLogic(){
  /* дополнительная логика (для с/у)
   
   0-канал - свет с/у
   1-канал - вытяжка с/у
   
   если 0 канал включен больше, чем 1.5 минуты (время установленное в MySwitch[0].timer), то нужно включить и 1 канал.
   после выключение 0 канала - 1 канал выключить через 5 минут (время установленное в MySwitch[1].timer)
   */

  // Если свет горит больше заданного интервала (параметр MySwitch[0].timer и он - ненулевой), а вытяжка не включена - нужно включить вытяжку
  if ((onFlag == false) && (millis() > (MySwitch[0].power_on + MySwitch[0].timer*60000)) && (MySwitch[0].timer != 0) && (MySwitch[0].state == HIGH) && (MySwitch[1].state == LOW) && (MySwitch[1].press_stop < MySwitch[0].power_on)) {
    // Включаем вытяжку
    doSwitch(1, HIGH);
    // Взводим флаг автовключения
    onFlag = true;
    logicFlag = true;
    // Выключаем режим автовыключения по таймауту
    MySwitch[1].autostate = false;
#ifndef DEBUGS
    Serial.println("Auto Logic ON");
#endif
  }

  // Если вытяжка включена - дадим ей новое время выключения - через заданный интервал ((параметр MySwitch[1].timer и он - ненулевой) после выключения света
  if ((logicFlag == true) && (offFlag == false) && (MySwitch[1].timer != 0) && (MySwitch[1].state == HIGH) && (MySwitch[0].state == LOW)) {
    MySwitch[1].time_off = millis() + MySwitch[1].timer*60000;
    MySwitch[1].autostate = true;
    offFlag = true;
#ifndef DEBUGS
    Serial.println("Auto Logic OFF started");
#endif
  }

  // Если все выключено, сбрасываем все флаги
  if ((logicFlag == true) && (MySwitch[0].state == LOW) && (MySwitch[1].state == LOW)) {
    offFlag = false;
    onFlag = false;
    logicFlag = false;

#ifndef DEBUGS
    Serial.println("Logic reset");
#endif
  }

  // Если при включенной вытяжке выключатель выключили вручную - запускаем автовыключение вытяжки
  if ((logicFlag == false) && (offFlag == false) && (MySwitch[0].press_stop > MySwitch[1].power_on) && (MySwitch[1].state == HIGH) && (MySwitch[0].state == LOW)) {
    logicFlag = true;
#ifndef DEBUGS
    Serial.println("Auto OFF 1 after manual OFF 0");
#endif
  }
}

// Функция обработки входящих сообщений от шлюза
void incomingMessage(const MyMessage &message)
  {
    switch (message.type) { // Разделение входящих сообщений по типу датчика (V_STATUS, V_WEIGHT, V_FLOW)
  case V_STATUS:  // Если сообщение для датчика реле
   if (message.sensor < CH)  // Проверить, действительно ли сообщение для реле
    {
      // Установить входящее сообщение в параметр состояние
      MySwitch[message.sensor].state = message.getBool(); // getBool - отдает только 0 или 1
      // Отправить в функцию выключателя, номер канала и состояние
      doSwitch(message.sensor, MySwitch[message.sensor].state);
      }
    break;
  case V_WEIGHT:  // Если сообщение для таймеров "auto_off"
      // Установить входящее сообщение в параметр таймера
      MySwitch[message.sensor-CH].auto_off = message.getByte();  // getByte - отдает числа 0-255
      // Сохранить состояние датчика в EEPROM (положение == номер датчика)
      gw.saveState( message.sensor, MySwitch[message.sensor-CH].auto_off);
      // Обновим параметры канала
      //MySwitch[message.sensor-CH].auto_off = ((unsigned long)MySwitch[message.sensor].auto_off);
#ifndef DEBUGS
      Serial.print("Off time is set ch");
      Serial.println(message.sensor - CH);
      Serial.print(MySwitch[message.sensor-CH].auto_off);
      Serial.println(" min");
#endif
    break;
  case V_FLOW:  // Если сообщение для таймеров "timer"
      // Установить входящее сообщение в параметр таймера
      MySwitch[message.sensor-CH-CH].timer = message.getByte();  // getByte - отдает числа 0-255
      // Сохранить состояние датчика в EEPROM (положение == номер датчика)
      gw.saveState( message.sensor, MySwitch[message.sensor-CH-CH].timer);
      // Обновим параметры канала
      //MySwitch[message.sensor-CH-CH].timer = ((unsigned long)MySwitch[message.sensor].timer);
#ifndef DEBUGS
      Serial.print("Timer ");
      Serial.println(message.sensor - CH - CH);
      Serial.print(MySwitch[message.sensor-CH-CH].timer);
      Serial.println(" min");
#endif
    break;
  default:
    break; 
    }
  }

